/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbsJmxServiceImpl.java 10798 2007-07-01 20:56:02Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.jmx;

import java.util.List;

import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MBeanServerConnection;
import javax.management.MBeanServerFactory;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.NamingException;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.apache.commons.modeler.Registry;


/**
 * JMX Service implementation. This singleton class must exist in each jonas
 * server that is to be administered via JMX. Its main role is to initialize the
 * service (the singleton object).
 * @author Guillaume Riviere
 * @author Michel Bruno
 * @author Adriana Danes
 */
public abstract class AbsJmxServiceImpl extends AbsServiceImpl implements JmxService {

    /**
     * Trace logger.
     */
    private static Logger logger = null;

    /**
     * @return Logger logger object
     */
    protected static Logger getLogger() {
        return logger;
    }

    /**
     * JMX server (MBean server) instance.
     */
    private MBeanServer jmxServer = null;

    /**
     * Jakarta commons modeler registry.
     */
    private Registry modelerRegistry = null;

    /**
     * Static description of all the packages that contain mbean descriptors.
     */
    private static final String[] PACKAGE_DESCRIPTORS = {
        "org.objectweb.jonas.server" // J2EEServer, J2EEDomain, JVM, WorkManager (which is not registered
                                     // by the Server but by the EJBService, and should also be registered
                                     // by the ResourceService
        , "org.objectweb.jonas.resource"
        , "org.ow2.jonas.web.tomcat6"};

    /**
     * Init the Service. Create the MBeanServer. Configuration information is
     * passed through a Context object.
     * @param ctx context containing service initilaisation parameters
     * @throws ServiceException the service could not be initialized
     */
    public void doInit(final Context ctx) throws ServiceException {
        logger = Log.getLogger(Log.JONAS_JMX_PREFIX);
        super.initLogger(Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX));

        // Create the MBeanServer

        // get Id
        String idMBeanServer = null;
        try {
            idMBeanServer = (String) ctx.lookup("idMBeanServer");
        } catch (NamingException ne) {
            throw new ServiceException("Cannot found the id of the MBean Server", ne);
        }

        // ----------------------
        List mbeanServers = MBeanServerFactory.findMBeanServer(idMBeanServer);
        if (mbeanServers.size() == 0) {
            throw new ServiceException("No MBean server found for id '" + idMBeanServer + "'.");
        }
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Use first MBean server ID found");
        }
        jmxServer = (MBeanServer) mbeanServers.get(0);

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "JMX Service initialized");
        }

        modelerRegistry = getModelRegistry();
    }

    /**
     * Start the Service Initialization of the service is already done.
     * @throws ServiceException the service could not be started
     */
    public abstract void doStart() throws ServiceException;

    /**
     * Stop this service
     * @throws ServiceException the service could not be stopped
     */
    public abstract void doStop() throws ServiceException;

    /**
     * @return The local reference of the MBean server
     */
    public MBeanServer getJmxServer() {
        return this.jmxServer;
    }

    /**
     * @return The Connection interface to the MBean server
     */
    public MBeanServerConnection getJmxServerConnection() {
        return (MBeanServerConnection) this.jmxServer;
    }

    /**
     * Register an MBean in the JOnAS MBeanServer
     * @param mbean MBean object to register
     * @param objectName the String representation of the MBean's ObjectName
     */
    public void registerMBean(Object mbean, String objectName) {
        if (jmxServer == null) {
            logger.log(BasicLevel.ERROR, "JMX Server not yet initialized");
            return;
        }
        try {
            ObjectName on = ObjectName.getInstance(objectName);
            jmxServer.registerMBean(mbean, on);
        } catch (MalformedObjectNameException e) {
            logger.log(BasicLevel.ERROR, objectName + ":" + e);
        } catch (InstanceAlreadyExistsException e) {
            logger.log(BasicLevel.WARN, objectName + ":" + e);
        } catch (MBeanRegistrationException e) {
            logger.log(BasicLevel.ERROR, objectName + ":" + e);
        } catch (NotCompliantMBeanException e) {
            logger.log(BasicLevel.ERROR, objectName + ":" + e);
        }
    }

    /**
     * Register an MBean in the JOnAS MBeanServer
     * @param mbean MBean object to register
     * @param objectName the MBean's ObjectName
     */
    public void registerMBean(Object mbean, ObjectName objectName) {
        if (jmxServer == null) {
            logger.log(BasicLevel.ERROR, "JMX Server not yet initialized");
            return;
        }
        try {
            jmxServer.registerMBean(mbean, objectName);
        } catch (InstanceAlreadyExistsException e) {
            logger.log(BasicLevel.WARN, objectName + ":" + e);
        } catch (MBeanRegistrationException e) {
            logger.log(BasicLevel.ERROR, objectName + ":" + e);
        } catch (NotCompliantMBeanException e) {
            logger.log(BasicLevel.ERROR, objectName + ":" + e);
        }
    }

    /**
     * Unregister an MBean from the JOnAS MBeanServer
     * @param objectName the MBean's ObjectName
     */
    public void unregisterMBean(ObjectName objectName) {
        if (jmxServer == null) {
            logger.log(BasicLevel.ERROR, "JMX Server not yet initialized");
            return;
        }
        try {
            jmxServer.unregisterMBean(objectName);
        } catch (InstanceNotFoundException e) {
            logger.log(BasicLevel.ERROR, objectName + ":" + e);
        } catch (MBeanRegistrationException e) {
            logger.log(BasicLevel.ERROR, objectName + ":" + e);
        }
    }

    /**
     * Register a Model MBean in the JOnAS MBeanServer.
     * @param mbean MBean object to register
     * @param objectName the Mbean ObjectName
     * @throws Exception throwed when registering a modeler MBean
     */
    public void registerModelMBean(Object mbean, ObjectName objectName) throws Exception {
        if (modelerRegistry == null) {
            logger.log(BasicLevel.ERROR, "Modeler MBeans Registry not yet created");
            return;
        }
        modelerRegistry.registerComponent(mbean, objectName, null);
    }

    /**
     * Register a Model MBean in the JOnAS MBeanServer.
     * @param mbean MBean object to register
     * @param objectName the Mbean stringified ObjectName
     * @throws Exception throwed when registering a modeler MBean
     */
    public void registerModelMBean(Object mbean, String objectName) throws Exception {
        registerModelMBean(mbean, ObjectName.getInstance(objectName));
    }

    /**
     * Unegister a Model MBean from the JOnAS MBeanServer
     * @param objectName the Mbean ObjectName
     */
    public void unregisterModelMBean(ObjectName objectName) {
        if (modelerRegistry == null) {
            logger.log(BasicLevel.ERROR, "Modeler MBeans Registry not yet created");
            return;
        }
        modelerRegistry.unregisterComponent(objectName);
    }

    /**
     * Remove internal references to the MBeanServer.
     */
    protected void releaseJmxServer() {
        try {
            MBeanServerFactory.releaseMBeanServer(jmxServer);
        } catch (java.lang.IllegalArgumentException ex) {
            // The mbeanServer is not found.
            logger.log(BasicLevel.ERROR, "Cannot find the MBeanServer" + ex);
        }
    }

    /**
     * Create Registry for model MBeans and load known descriptors
     * @return the created commens modeler registry
     */
    private Registry getModelRegistry() {
        // Load registry
        Registry registry = Registry.getRegistry(null, null);
        ClassLoader cl = AbsJmxServiceImpl.class.getClassLoader();

        // Load descriptors
        for (int i = 0; i < PACKAGE_DESCRIPTORS.length; i++) {
            registry.loadDescriptors(PACKAGE_DESCRIPTORS[i], cl);
        }

        // Log loaded descriptors
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            String s;
            String[] as = registry.findManagedBeans();
            logger.log(BasicLevel.DEBUG, ">>> List of all MBeans descriptors");
            for (int i = 0; i < as.length; i++) {
                s = ">>> " + i + ") " + as[i];
                logger.log(BasicLevel.DEBUG, s);
            }
            logger.log(BasicLevel.DEBUG, "<<< List of all MBeans descriptors");
        }
        return (registry);
    }

    /**
     * Load additional mbean descriptors.
     * @param packageName name of the package containing the descriptors file
     * @param cl class loader containing the resource
     */
    public void loadDescriptors(String packageName, ClassLoader cl) {
        if (cl == null) {
            cl = AbsJmxServiceImpl.class.getClassLoader();
        }
        modelerRegistry.loadDescriptors(packageName, cl);
    }
}